"""
Async version of the base classes.
"""

import inspect
from abc import ABC, abstractmethod
from types import TracebackType
from typing import (
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    Generic,
    Iterable,
    MutableSequence,
    MutableSet,
    Optional,
    Self,
    Sequence,
    Type,
    cast,
    overload,
)
from warnings import warn

from ._sl import _SL
from .exceptions import EntityNotFound, ImplicitRollback, LogicError, TooManyRowsError
from .generics import OBJ, OBJ_OR_FACTORY, ConnectionTypeT, DictT, ListT, ResObjT, ResultRow, SetT, T


class Result(ABC):
    """
    Asynchronous result of query
    """

    @overload
    async def fetch_list(self, /) -> Optional[list[Any]]: ...

    @overload
    async def fetch_list(self, cls: Type[ListT], /, *args: Any, **kwargs: Any) -> Optional[ListT]: ...

    @abstractmethod
    async def fetch_list(self, cls: Type[ListT | list[Any]] = list, /, *args: Any, **kwargs: Any) -> Optional[ListT | list[Any]]:
        """
        Fetch one row from result as list. The list type can be specified as the cls argument.
        :param cls: List type that should be returned. Defaults to plain list.
        :param args: Optional arguments to the cls constructor.
        :param kwargs: Optional kwargs to the cls constructor.
        :return: Instance of cls created from result row or None if no rows are remaining in the result set.
        """

    @overload
    async def fetch_dict(self, /) -> Optional[dict[str, Any]]:
        """
        Fetch one row from result as dictionary. The dictionary type can be specified as the cls argument.
        :return: dict containing mapping from column names to values or None if no rows are remaining in the result set.
        """

    @overload
    async def fetch_dict(self, cls: Type[DictT], /, *args: Any, **kwargs: Any) -> Optional[DictT]:
        """
        Fetch one row from result as dictionary. The dictionary type can be specified as the cls argument.
        :param cls: Dictionary type that should be returned.
        :param args: Optional arguments to the cls constructor.
        :param kwargs: Optional kwargs for the constructor.
        :return: Instance of cls created from result row or None if no rows are remaining in the result set.
        """

    @abstractmethod
    async def fetch_dict(
        self, cls: Type[DictT | dict[str, Any]] = dict[str, Any], /, *args: Any, **kwargs: Any
    ) -> Optional[DictT | dict[str, Any]]:
        """
        Fetch one row from result as dictionary. The dictionary type can be specified as the cls argument.
        :param cls: Dictionary type that should be returned. Defaults to plain dict.
        :param args: Optional arguments to the cls constructor.
        :param kwargs: Optional kwargs for the constructor.
        :return: Instance of cls created from result row or None if no rows are remaining in the result set.
        """

    @overload
    async def fetch_object(self, /) -> Optional[ResultRow]: ...

    @overload
    async def fetch_object(self, cls: OBJ_OR_FACTORY[OBJ], /, *args: Any, **kwargs: Any) -> Optional[OBJ]: ...

    @abstractmethod
    async def fetch_object(
        self, cls: OBJ_OR_FACTORY[OBJ] | Type[ResultRow] = ResultRow, /, *args: Any, **kwargs: Any
    ) -> Optional[OBJ | ResultRow]:
        """
        Fetch one row from result as object of type specified by the cls argument.
        :param cls: Class (or factory) representing the object that should be returned.
        :param args: Arguments to the cls constructor.
        :param as_kwargs: Should the row attributes be passed as kwargs to the constructor? If False, properties
          will be set directly using setattr().
        :param kwargs: Optional kwargs for the constructor.
        :return: Instance of cls created from result row or None if no rows are remaining in the result set.
        """

    @staticmethod
    async def _fetch_all(
        method: Callable[[Any, Any, Any], Awaitable[Any]], cls: Type[OBJ] | Callable[..., OBJ], *args: Any, **kwargs: Any
    ) -> AsyncGenerator[OBJ, None]:
        while True:
            row = await method(cls, *args, **kwargs)

            if row is None:
                break

            yield row

    @overload
    def fetch_all_object(self, /) -> AsyncGenerator[ResultRow, None]: ...

    @overload
    def fetch_all_object(self, cls: OBJ_OR_FACTORY[OBJ], /, *args: Any, **kwargs: Any) -> AsyncGenerator[OBJ, None]: ...

    async def fetch_all_object(
        self, cls: OBJ_OR_FACTORY[OBJ] | Type[ResultRow] = ResultRow, /, *args: Any, **kwargs: Any
    ) -> AsyncGenerator[ResultRow | OBJ, None]:
        """
        Generate objects of type cls for each row of the result set, with optionally passing args and kwargs to the
        cls constructor.
        :param cls: See fetch_object for details
        :param args: See fetch_object for details
        :param kwargs: See fetch_object for details
        :return: Generator generating result rows as objects
        """
        async for row in self._fetch_all(self.fetch_object, cls, *args, **kwargs):
            yield cast(OBJ, row)

    @overload
    def fetch_all_dict(self, /) -> AsyncGenerator[dict[str, Any], None]: ...

    @overload
    def fetch_all_dict(self, cls: Type[DictT], /, *args: Any, **kwargs: Any) -> AsyncGenerator[DictT, None]: ...

    async def fetch_all_dict(
        self, cls: Type[DictT | dict[str, Any]] = dict[str, Any], /, *args: Any, **kwargs: Any
    ) -> AsyncGenerator[DictT | dict[str, Any], None]:
        """
        Generate dicts of type cls for each row of the result set, with optionally passing args and kwargs to the
        cls constructor.
        :param cls: See fetch_dict for details
        :param args: See fetch_dict for details
        :param kwargs: See fetch_dict for details
        :return: Generator generating result rows as dicts
        """
        async for row in self._fetch_all(self.fetch_dict, cls, *args, **kwargs):
            yield cast(DictT, row)

    @overload
    def fetch_all_list(self, /) -> AsyncGenerator[list[Any], None]: ...

    @overload
    def fetch_all_list(self, cls: Type[ListT], /, *args: Any, **kwargs: Any) -> AsyncGenerator[ListT, None]: ...

    async def fetch_all_list(
        self, cls: Type[ListT | list[Any]] = list, /, *args: Any, **kwargs: Any
    ) -> AsyncGenerator[ListT | list[Any], None]:
        """
        Generate lists of type cls for each row of the result set, with optionally passing args and kwargs to the
        cls constructor.
        :param cls: See fetch_list for details
        :param args: See fetch_list for details
        :param kwargs: See fetch_list for details
        :return: Generator generating result rows as lists
        """
        async for row in self._fetch_all(self.fetch_list, cls, *args, **kwargs):
            yield cast(ListT, row)

    @overload
    async def fetch_all_scalar(self) -> list[Any]:
        """
        Fetch all rows from result as a list of scalars. Only for queries returning single column.
        """

    @overload
    async def fetch_all_scalar(self, factory: Type[SetT]) -> SetT:
        """
        Fetch all rows from result as a list of scalars. Only for queries returning single column.
        :param factory: Factory to use for the output type.
        """

    @overload
    async def fetch_all_scalar(self, factory: Type[ListT]) -> ListT:
        """
        Fetch all rows from result as a list of scalars. Only for queries returning single column.
        :param factory: Factory to use for the output type.
        """

    async def fetch_all_scalar(self, factory: Type[SetT | ListT | list[Any]] = list) -> SetT | ListT | list[Any]:
        """
        Fetch all rows from the result set as scalars.
        :param factory: Factory to use for the output type. Can be MutableSet or MutableSequence.
        """
        if not inspect.isclass(factory) or not issubclass(factory, (MutableSet, MutableSequence)):
            raise AttributeError("Unknown type factory. Can be only MutableSet or MutableSequence.")

        out = factory()
        append = cast(ListT, out).append if issubclass(factory, MutableSequence) else cast(SetT, out).add

        async for row in self.fetch_all_list():
            append(row[0])

        return cast(SetT | ListT, out)


class Transaction(ABC, Generic[ConnectionTypeT]):
    """
    Asynchronous transaction
    """

    def __init__(self, connection: ConnectionTypeT) -> None:
        self._connection = connection

    def __enter__(self) -> None:
        """
        Asynchronous transaction cannot be used in synchronous context.
        """
        raise RuntimeError("Cannot use async transaction in sync context. Use async with ...")

    async def __aenter__(self) -> Self:
        """
        Enter asynchronous transaction context.
        """
        return self

    async def __aexit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> None:
        """
        If transaction is not explicitly committed or rollbacked, it is rollbacked as cleanup strategy.
        """
        if not await self.is_committed():
            warn("Rolling back uncommitted transaction.", ImplicitRollback, stacklevel=_SL.get() + 2)
            await self.rollback()

    @abstractmethod
    async def is_committed(self) -> bool:
        """
        Return True if transaction has been commited.
        """

    @abstractmethod
    async def last_insert_id(self) -> int:
        """
        Return last used AUTO_INCREMENT value.
        """

    @abstractmethod
    async def query(self, query: str, *args: Any, **kwargs: Any) -> Result | bool:
        """
        Execute single query inside the transaction scope.
        :param query: Query to execute. Can contain %s for each arg from args.
        :param args: Arguments to the query
        :param kwargs: Some drivers supports also specifying keyword arguments for the query, but most does not.
        :return: Result if query returns rows, or True if query succeeded.
        :raises: QueryError (or subclass of) if there was an error processing the query.
        """

    @abstractmethod
    async def affected_rows(self) -> int:
        """
        Return number of affected rows by most recent query.
        """

    @abstractmethod
    async def commit(self) -> None:
        """
        Commit the transaction. After commit, no other operation can be performed on the transaction.
        :raises: QueryError (or subclass of) if there was an error during commit.
        """

    @abstractmethod
    async def rollback(self) -> None:
        """
        Rollback the transaction. After rollback, no other operation can be performed on the transaction.
        :raises: QueryError (or subclass of) if there was an error during rollback.
        """

    async def _get(
        self,
        result_fetcher: Callable[..., Callable[..., Awaitable[Any]]],
        query: str,
        args: Optional[Iterable[Any]],
        cls: Type[T] | Callable[..., T],
        *cls_args: Any,
        **cls_kwargs: Any,
    ) -> T:
        res = await self.query(query, *(args or []))
        if not isinstance(res, Result):
            raise LogicError(query, "Query did not return result.")

        row = await result_fetcher(res)(cls, *cls_args, **cls_kwargs)
        if row is None:
            raise EntityNotFound(query, "Requested entity was not found.")

        another = await result_fetcher(res)(cls, *cls_args, **cls_kwargs)
        if another is not None:
            raise TooManyRowsError(query, "Query returned more than one row.")

        return row

    @overload
    async def get_object(self, query: str, args: Optional[Iterable[Any]] = None, /) -> ResultRow: ...

    @overload
    async def get_object(
        self, query: str, args: Optional[Iterable[Any]] = None, cls: Type[OBJ] = ..., /, *cls_args: Any, **cls_kwargs: Any
    ) -> OBJ: ...

    @overload
    async def get_object(
        self,
        query: str,
        args: Optional[Iterable[Any]] = None,
        cls: Callable[..., OBJ] = ...,
        /,
        *cls_args: Any,
        **cls_kwargs: Any,
    ) -> OBJ: ...

    async def get_object(
        self,
        query: str,
        args: Optional[Iterable[Any]] = None,
        cls: OBJ_OR_FACTORY[OBJ] | Type[ResultRow] = ResultRow,
        /,
        *cls_args: Any,
        **cls_kwargs: Any,
    ) -> OBJ | ResultRow:
        """
        Retrieve single object from the database. The query must return exactly one row, which will be
        converted to object using standard Result.fetch_object.
        :param query: Query to execute
        :param args: Arguments to the query
        :param cls: See Result.fetch_object for details
        :param cls_args: See Result.fetch_object for details
        :param cls_kwargs: See Result.fetch_object for details
        :return: Object retrieved by the query.
        :raises: LogicError if query did not return a result set (i.e. modifying query instead of SELECT).
        :raises: EntityNotFound if query returned a result set, but it was empty.
        :raises: TooManyRowsError if query returned more than single row.
        """
        return await self._get(lambda res: res.fetch_object, query, args, cls, *cls_args, **cls_kwargs)

    @overload
    async def get_dict(self, query: str, args: Optional[Iterable[Any]] = None, /) -> dict[str, Any]: ...

    @overload
    async def get_dict(
        self, query: str, args: Optional[Iterable[Any]] = None, cls: Type[DictT] = ..., /, *cls_args: Any, **cls_kwargs: Any
    ) -> DictT: ...

    async def get_dict(
        self,
        query: str,
        args: Optional[Iterable[Any]] = None,
        cls: Type[DictT | dict[str, Any]] = dict,
        /,
        *cls_args: Any,
        **cls_kwargs: Any,
    ) -> DictT | dict[str, Any]:
        """
        Retrieve single row from the database as dict. The query must return exactly one row, which will be
        converted to dict using standard Result.fetch_dict.
        :param query: Query to execute
        :param args: Arguments to the query
        :param cls: See Result.fetch_dict for details
        :param cls_args: See Result.fetch_dict for details
        :param cls_kwargs: See Result.fetch_dict for details
        :return: Dict retrieved by the query.
        :raises: LogicError if query did not return a result set (i.e. modifying query instead of SELECT).
        :raises: EntityNotFound if query returned a result set, but it was empty.
        :raises: TooManyRowsError if query returned more than single row.
        """
        return await self._get(lambda res: res.fetch_dict, query, args, cls, *cls_args, **cls_kwargs)

    @overload
    async def get_list(self, query: str, args: Optional[Iterable[Any]] = None, /) -> list[Any]: ...

    @overload
    async def get_list(
        self, query: str, args: Optional[Iterable[Any]] = None, cls: Type[ListT] = ..., /, *cls_args: Any, **cls_kwargs: Any
    ) -> ListT: ...

    async def get_list(
        self,
        query: str,
        args: Optional[Iterable[Any]] = None,
        cls: Type[ListT | list[Any]] = list,
        /,
        *cls_args: Any,
        **cls_kwargs: Any,
    ) -> ListT | list[Any]:
        """
        Retrieve single row from the database as list. The query must return exactly one row, which will be
        converted to list using standard Result.fetch_list.
        :param query: Query to execute
        :param args: Arguments to the query
        :param cls: See Result.fetch_list for details
        :param cls_args: See Result.fetch_list for details
        :param cls_kwargs: See Result.fetch_list for details
        :return: List retrieved by the query.
        :raises: LogicError if query did not return a result set (i.e. modifying query instead of SELECT).
        :raises: EntityNotFound if query returned a result set, but it was empty.
        :raises: TooManyRowsError if query returned more than single row.
        """
        return await self._get(lambda res: res.fetch_list, query, args, cls, *cls_args, **cls_kwargs)

    async def get_scalar(self, query: str, args: Optional[Iterable[Any]] = None) -> Any:
        """
        Retrieve single value from the database. The query must return exactly one row and that
        row must contain exactly one column, which is then returned from this method.
        :param query: Query to execute
        :param args: Query arguments.
        :return: Value of first column of the first row of the result set.
        :raises: LogicError if query did not return a result set (i.e. modifying query instead of SELECT).
        :raises: EntityNotFound if query returned a result set, but it was empty.
        :raises: TooManyRowsError if query returned more than single row or a row contains more than single column.
        """
        row: Sequence[Any] = await self.get_list(query, args)
        if len(row) != 1:
            raise TooManyRowsError(query, "Result must contain only one column to be fetched as scalar.")

        return row[0]

    @property
    @abstractmethod
    def connection(self) -> ConnectionTypeT:
        """
        Return connection that this transaction is open on.
        :return: connection object
        """

    @abstractmethod
    async def cleanup(self) -> None:
        """
        Cleanup the connection associated with this transaction and prepare it for usage by next transaction.
        """


class ObjectFetcher(Generic[ResObjT]):
    """
    Helper class that acts as generator that fetches objects. The only difference between Result.fetch_all_object is,
    that this can be type-safe parametrized with actual type of object that it should generate, which helps with
    code completion and mypy checks.
    """

    def __init__(
        self,
        q: Result,
        cls: Type[ResultRow | ResObjT] | Callable[..., ResObjT] = ResultRow,
        cls_args: Optional[Iterable[Any]] = None,
        cls_kwargs: Optional[dict[str, Any]] = None,
        as_kwargs: bool = True,
    ):
        """
        Construct the ObjectFetcher.
        :param q: Query result
        :param cls: See Result.fetch_object for details.
        :param cls_args: See Result.fetch_object for details.
        :param cls_kwargs: See Result.fetch_object for details.
        :param as_kwargs: See Result.fetch_object for details.
        """
        self.q: Result = q
        self.cls = cls
        self.cls_args = cls_args or []
        self.cls_kwargs = cls_kwargs or {}
        self.as_kwargs = as_kwargs

    def __aiter__(self) -> Self:
        """
        Begining of the iteration.
        """
        return self

    async def __anext__(self) -> ResObjT:
        """
        Return one row from the result set as object of specified type.
        """
        resp = cast(ResObjT, await self.q.fetch_object(self.cls, *self.cls_args, as_kwargs=self.as_kwargs, **self.cls_kwargs))
        if resp is None:
            raise StopAsyncIteration()

        return resp
